import urllib.parse

from .exceptions import ZivverMissingRequiredFields, ZivverCRUDError
from .external_connection import OauthConnection
from .wrapper import get_zivver_user_object


class ZivverSCIMConnection:
    """
    Object representing the zivver connection to do CRUD operations with
    """

    def __init__(self, external_oauth_token_value, scim_api_create_url, scim_api_update_url,
                 scim_api_get_url, scim_api_delete_url):
        self.external_oauth_token_value = external_oauth_token_value
        self.scim_api_create_url = scim_api_create_url
        self.scim_api_update_url = scim_api_update_url
        self.scim_api_get_url = scim_api_get_url
        self.scim_api_delete_url = scim_api_delete_url

    def _check_required_create_fields(self, last_name=None, user_name=None, sso_connection=None,
                               zivver_account_key=None):
        # Check for required fields
        if not self.external_oauth_token_value:
            raise ZivverMissingRequiredFields('Missing field: external_oauth_token_value')
        if not self.scim_api_create_url:
            raise ZivverMissingRequiredFields('Missing field: scim_api_create_url')
        if not self.scim_api_update_url:
            raise ZivverMissingRequiredFields('Missing field: scim_api_update_url')
        if not self.scim_api_get_url:
            raise ZivverMissingRequiredFields('Missing field: scim_api_get_url')
        if not self.scim_api_delete_url:
            raise ZivverMissingRequiredFields('Missing field: scim_api_delete_url')

        if not last_name:
            raise ZivverMissingRequiredFields('Missing field: last_name')
        if not user_name:
            raise ZivverMissingRequiredFields('Missing field: user_name')

        if sso_connection and not zivver_account_key:
            raise ZivverMissingRequiredFields('Missing field: zivver_account_key')

    def _check_required_delete_get_fields(self, account_id):
        """
        Checks if the account ID is empty
        :param account_id:
        :except If the required account_id is empty
        """
        if not account_id:
            raise ZivverMissingRequiredFields('Missing field: account_id')

    def _check_response(self, response):
        """
        Check the repsone for errors, raise if there are any errors.
        """
        if type(response) is not dict and response.status_code in [400, 401, 403, 404]:
            raise ZivverCRUDError(message='Response from Zivver with Errors', response=response)

    def create_user_in_zivver(self, first_name=None, last_name=None, nick_name=None, user_name=None,
                              zivver_account_key=None, sso_connection=False, is_active=False, aliases=[],
                              delegates=[]):
        """
        Create a user in Zivver via SCIM.
        :return: Returns the ZivverUser() object
        """
        self._check_required_create_fields(last_name=last_name, user_name=user_name, sso_connection=sso_connection,
                                           zivver_account_key=zivver_account_key)

        # Set defaults
        if not first_name:
            first_name = ''

        scim_object_user = {
            'schemas': [
                'urn:ietf:params:scim:schemas:core:2.0:User',
                'urn:ietf:params:scim:schemas:extension:enterprise:2.0:User',
                'urn:ietf:params:scim:schemas:zivver:0.1:User'
            ],
            'meta': {
                'resourceType': 'User'
            },
            'active': is_active,
            'name': {
                'formatted': '{} {}'.format(first_name, last_name)
            },
            'nickName': nick_name,
            'urn:ietf:params:scim:schemas:zivver:0.1:User': {
                'SsoAccountKey': zivver_account_key,
                'aliases': aliases,
                'delegates': delegates
            },
            'userName': user_name
        }

        oauth_connection = OauthConnection(external_oauth_token_value=self.external_oauth_token_value)
        response = oauth_connection.return_request_post_data(post_url=self.scim_api_create_url,
                                                             object_serialized=scim_object_user)

        self._check_response(response)

        zivver_user = get_zivver_user_object(response)
        return zivver_user

    def delete_user_from_zivver(self, account_id):
        """
        Delete the user from Zivver.
        NOTE: Deleting the user is irreversible
        :return: Returns the response object
        """
        self._check_required_delete_get_fields(account_id)

        oauth_connection = OauthConnection(external_oauth_token_value=self.external_oauth_token_value)
        delete_url = urllib.parse.urljoin(self.scim_api_delete_url, account_id)
        response = oauth_connection.return_request_delete_data(delete_url=delete_url)

        self._check_response(response)

        return response

    def get_user_from_zivver(self, account_id):
        """
        Returns the user from Zivver if the user exists
        :param account_id:
        :return: ZivverUser() object
        """
        self._check_required_delete_get_fields(account_id)

        oauth_connection = OauthConnection(external_oauth_token_value=self.external_oauth_token_value)
        get_url = urllib.parse.urljoin(self.scim_api_get_url, account_id)
        response = oauth_connection.return_request_get_data(get_url=get_url)

        self._check_response(response)

        zivver_user = get_zivver_user_object(response)
        return zivver_user

    def get_all_users_from_zivver(self):
        """
        Returns a list of users from Zivver
        :param account_id:
        :return: List(ZivverUser()) object
        """
        oauth_connection = OauthConnection(external_oauth_token_value=self.external_oauth_token_value)
        response = oauth_connection.return_request_get_data(get_url=self.scim_api_get_url)

        self._check_response(response)

        zivver_users = []
        for zivver_scim_user in response['Resources']:
            zivver_users.append(get_zivver_user_object(zivver_scim_user))

        return zivver_users

    def update_user_in_zivver(self, account_id, first_name=None, last_name=None, nick_name=None, user_name=None,
                              zivver_account_key=None, sso_connection=False, is_active=False, aliases=[],
                              delegates=[]):
        """
        Update a user in Zivver via SCIM.
        :return: Returns the ZivverUser() object
        """
        self._check_required_create_fields(last_name=last_name, user_name=user_name, sso_connection=sso_connection,
                                           zivver_account_key=zivver_account_key)

        self._check_required_delete_get_fields(account_id)

        # Set defaults
        if not first_name:
            first_name = ''

        scim_object_user = {
            'schemas': [
                'urn:ietf:params:scim:schemas:core:2.0:User',
                'urn:ietf:params:scim:schemas:extension:enterprise:2.0:User',
                'urn:ietf:params:scim:schemas:zivver:0.1:User'
            ],
            'meta': {
                'resourceType': 'User',
                'location': '/scim/v2/Users/{}'.format(account_id)
            },
            'active': is_active,
            'id': account_id,
            'name': {
                'formatted': '{} {}'.format(first_name, last_name)
            },
            'nickName': nick_name,
            'urn:ietf:params:scim:schemas:zivver:0.1:User': {
                'SsoAccountKey': zivver_account_key,
                'aliases': aliases,
                'delegates': delegates
            },
            'userName': user_name
        }

        oauth_connection = OauthConnection(external_oauth_token_value=self.external_oauth_token_value)
        put_url = urllib.parse.urljoin(self.scim_api_update_url, account_id)
        response = oauth_connection.return_request_put_data(put_url=put_url, object_serialized=scim_object_user)

        self._check_response(response)

        zivver_user = get_zivver_user_object(response)
        return zivver_user



import json
import requests


class OauthConnection:
    """
    Object will be post/update/delete via this class in the external application
    """

    def __init__(self, external_oauth_token_value=None, extra_headers=None):
        self.external_oauth_token_value = external_oauth_token_value
        self.custom_oauth_header = {
            'header_key': 'Authorization',
            'header_value': 'Bearer {}'.format(self.external_oauth_token_value)
        }
        self.extra_headers = extra_headers

    def add_extra_headers(self, extra_headers):
        """
        Add extra headers to the OAuth object
        """
        self.extra_headers = extra_headers

    def set_custom_oauth_authorization_header(self, header_key, header_value):
        """
        Update the oauth Authorization header with your own key/value
        Default is:
        { 'Authorization': 'Bearer [token]',
        ...
        }
        """
        self.custom_oauth_header = {
            'header_key': header_key,
            'header_value': header_value
        }

    def _get_content_length_of_object_serialized(self, object_serialized):
        """
        Returns the length of the serialized object
        """
        content_length = len(json.dumps(object_serialized))
        return content_length

    def _create_authorization_header(self, object_serialized=None):
        """
        Creates the headers that are send with the requests
        """
        header_key = self.custom_oauth_header['header_key']
        header_value = self.custom_oauth_header['header_value']

        headers = {
            header_key: header_value,
            'Content-type': 'application/json',
            'Accept': 'application/json'
        }

        # add extra headers if any:
        if self.extra_headers:
            for key, val in self.extra_headers.items():
                headers[key] = val

        if object_serialized is not None:
            content_length = self._get_content_length_of_object_serialized(object_serialized)
            headers['Content-Length'] = '{}'.format(content_length)

        return headers

    def return_request_post_data(self, post_url, object_serialized):
        """
        Do a POST request to the URL
        :return: The defualt json() object, if none, then returns the response object
        """
        headers = self._create_authorization_header(object_serialized)
        result = requests.post(post_url, headers=headers, json=object_serialized)
        try:
            return result.json()
        except Exception:
            return result

    def return_request_get_data(self, get_url):
        """
        Do a GET request to the url and return the data
        :return: The defualt json() object, if none, then returns the response object
        """
        headers = self._create_authorization_header()
        result = requests.get(get_url, headers=headers)
        try:
            return result.json()
        except Exception:
            return result

    def return_request_delete_data(self, delete_url):
        """
        Requests a DELETE method
        :return: The defualt json() object, if none, then returns the response object
        """
        headers = self._create_authorization_header()
        result = requests.delete(delete_url, headers=headers)
        try:
            return result.json()
        except Exception:
            return result

    def return_request_patch_data(self, patch_url, object_serialized):
        """
        Do a PATCH request to the URL
        :return: The defualt json() object, if none, then returns the response object
        """
        headers = self._create_authorization_header(object_serialized)
        result = requests.patch(patch_url, headers=headers, json=object_serialized)
        try:
            return result.json()
        except Exception:
            return result

    def return_request_put_data(self, put_url, object_serialized):
        """
        Do a PUT request to the URL
        :return: The defualt json() object, if none, then returns the response object
        """
        headers = self._create_authorization_header(object_serialized)
        result = requests.put(put_url, headers=headers, json=object_serialized)
        try:
            return result.json()
        except Exception:
            return result



class ZivverMissingRequiredFields(Exception):
    """When there are missing required fields missing"""
    pass


class ZivverCRUDError(Exception):
    """When there are erros in the response"""

    def __init__(self, message, response, code=None, params=None):
        """
        Exception raised only when there are CRUD errors from Zivver.
        Saves the response object inside the response variable
        """
        super().__init__(message, code, params)

        self.response = response
        self.status_code = response.status_code
        self.reason = response.reason
        self.text = response.text

    def get_error_message(self):
        """
        :return: formatted text error message
        """
        return 'Status code: {}, Reason: {}, Text: {}'.format(self.status_code, self.reason, self.text)

    def get_sollution(self):
        """
        Error messages come from Zivver:
        https://docs.zivver.com/en/admin/synctool/troubleshooting/synchronization-log-errors.html
        :return: Sollutions based on Zivver cause
        """
        if 'Error creating alias: Alias is already taken for' in self.text:
            return """
                Cause
                The message Alias is already taken means that the email alias is a separate 
                account on the Zivver platform, instead of being an email alias like in Exchange.
                
                Solution
                To keep the received Zivver messages accessible to the recipient, 
                email alias must be manually merged on our platform to the primary email address.
                
                On the Account page, find the Zivver account that is mentioned first in the 
                error and merge the account mentioned after "error creating alias: 
                Alias is already taken for "ZivverUID" in the error.
                How to merge two accounts\r\n"""
        if 'Unknown account with uuid:' in self.text:
            return """
                Account probably does not exist, did you create the account?\r\n"""
        if 'Account is outside the organization' == self.text:
            return """
                There are multiple possible causes for this error:
                
                Cause 1
                The email address mentioned exists outside your Zivver organization. If this is the case, then you will
                not find the account on the Account page when you search for the email address. If you do find the 
                account, then go to Cause 2 below.

                Solution
                Go the the Domain page.
                Adopt free accounts

                Cause 2
                The Synctool tries to convert a Zivver user account to a functional account or the other way around. 
                The Synctool can’t change account types.
                
                Solution
                Change the account type of the account. Use the table below for the recommended 
                Zivver account types for each type of mailbox.
                
                Mailbox type	    Recommended Zivver account type
                User mailbox        Normal account
                Shared mailbox      Functional account\r\n"""


import json


def get_zivver_user_object(zivver_scim):
    """
    Wrapper to wrap Zivver SCIM to Zivver Python Class
    :param zivver_scim: Returned from the Zivver response, contains the object on create/update
    :return: Zivver()
    """    
    account_id = zivver_scim.get('id', '')

    name_formatted = zivver_scim.get('name', '')
    if name_formatted:
        name_formatted = name_formatted.get('formatted', '')

    meta = zivver_scim.get('meta', '')
    meta_created_at = ''
    meta_location = ''
    meta_resource_type = ''
    if meta:
        meta_created_at = meta.get('created', '')
        meta_location = meta.get('location', '')
        meta_resource_type = meta.get('resourceType', '')
        
    phone_numbers = zivver_scim.get('phoneNumbers', [])
    
    user_name = zivver_scim.get('userName', '')
    nick_name = zivver_scim.get('nickName', '')
    is_active = zivver_scim.get('active', False)

    schemas = zivver_scim.get('schemas', [])
    enterprise_user = zivver_scim.get('urn:ietf:params:scim:schemas:extension:enterprise:2.0:User', '')
    
    zivver_scim_user = zivver_scim.get('urn:ietf:params:scim:schemas:zivver:0.1:User', '')
    zivver_scim_user_aliases = ''
    zivver_scim_user_delegates = ''
    if zivver_scim_user:
        zivver_scim_user_aliases = zivver_scim_user.get('aliases', [])
        zivver_scim_user_delegates = zivver_scim_user.get('delegates', [])
    
    return ZivverUser(account_id=account_id, name_formatted=name_formatted, meta_location=meta_location,
                      meta_created_at=meta_created_at, meta_resource_type=meta_resource_type,
                      phone_numbers=phone_numbers, user_name=user_name, nick_name=nick_name, is_active=is_active,
                      schemas=schemas, enterprise_user=enterprise_user,
                      zivver_scim_user_aliases=zivver_scim_user_aliases,
                      zivver_scim_user_delegates=zivver_scim_user_delegates)


class ZivverUser:
    """
    ZivverUser Class object created from the ZivverUser create/update response
    """

    def __init__(self, account_id=None, name_formatted=None, meta_created_at=None, meta_location=None,
                 meta_resource_type=None, phone_numbers=None, user_name=None, nick_name=None, is_active=False,
                 schemas=None, enterprise_user=None, zivver_scim_user_aliases=None, zivver_scim_user_delegates=None):
        self.account_id = account_id
        self.name_formatted = name_formatted
        self.meta_created_at = meta_created_at
        self.meta_location = meta_location
        self.meta_resource_type = meta_resource_type
        self.phone_numbers = phone_numbers
        self.user_name = user_name
        self.nick_name = nick_name
        self.is_active = is_active
        self.schemas = schemas
        self.enterprise_user = enterprise_user
        self.zivver_scim_user_aliases = zivver_scim_user_aliases
        self.zivver_scim_user_delegates = zivver_scim_user_delegates

    def __str__(self):
        """
        Representation of the SCIM user response in JSON
        :return: SCIM Response json object
        """
        return json.dumps(
            {
                'id': self.account_id,
                'name': {
                    'formatted': self.name_formatted
                },
                'meta': {
                    'created': self.meta_created_at,
                    'location': self.meta_location,
                    'resourceType': self.meta_resource_type
                },
                'phoneNumbers': self.phone_numbers,
                'schemas': self.schemas,
                'userName': self.user_name,
                'nickName': self.nick_name,
                'active': self.is_active,
                'urn:ietf:params:scim:schemas:extension:enterprise:2.0:User': self.enterprise_user,
                'urn:ietf:params:scim:schemas:zivver:0.1:User': {
                    'aliases': self.zivver_scim_user_aliases,
                    'delegates': self.zivver_scim_user_delegates
                }
            }, indent=4
        )
